package dk.brics.xact;

import java.util.LinkedHashSet;
import java.util.Set;

import dk.brics.misc.Origin;

/**
 * Element node.
 */
public class Element extends TempNode {
	
	private final String namespace;

	private final String localname;

	private final AttrNode firstattribute;
	
	private final XML firstchild;
	
	private final NamespaceDecl firstnsdecl;
	
	/**
	 * Constructs a new element node.
	 */
	public Element(String namespace, String localname, AttrNode firstattribute, XML firstchild, 
			NamespaceDecl firstnsdecl, XML nextsibling, Origin origin) { // TODO: also check that all namespaces are legal URIs?
		super(nextsibling, origin);
		if (namespace == null)
			namespace = "";
		while (firstchild != null && firstchild instanceof Text && ((Text)firstchild).getString().length() == 0) 
			firstchild = firstchild.getRealNextSibling();
		if (!StringTypes.isNCName(localname) || localname.startsWith("xmlns"))
			throw new XMLWellformednessException("illegal element name " + localname);
		Set<String> attrnames = new LinkedHashSet<String>();
		for (AttrNode n = firstattribute; n != null; n = n.getNextAttr()) {
			String an = expanded(n.getNamespace(), n.getLocalName());
			if (attrnames.contains(an))
				throw new XMLWellformednessException("duplicate attribute name " + an);
			attrnames.add(an);
		}
		this.namespace = namespace;
		this.localname = localname;
		this.firstattribute = firstattribute;
		this.firstchild = firstchild;
		this.firstnsdecl = firstnsdecl;
	}
	
	/**
	 * Constructs a new empty element node.
	 */
	public Element(String namespace, String localname) {
		this(namespace, localname, null, null, null, null, null);
	}

	/**
	 * Constructs a new empty element node.
	 */
	public Element(String namespace, String localname, Origin origin) {
		this(namespace, localname, null, null, null, null, origin);
	}

	/**
	 * Constructs a new empty element node.
	 */
	public Element(String localname) {
		this(null, localname, null, null, null, null, null);
	}

	/**
	 * Constructs a new empty element node.
	 */
	public Element(String localname, Origin origin) {
		this(null, localname, null, null, null, null, origin);
	}

	/**
	 * Returns the namespace URI.
	 */
	public String getNamespace() {
		return namespace;
	}

	/**
	 * Returns the local name.
	 */
	public String getLocalName() {
		return localname;
	}

	/**
	 * Returns the expanded name.
	 */
	public String getExpandedName() {
		return expanded(namespace, localname);
	}

	/**
	 * Returns the first attribute-like node.
	 */
	public AttrNode getFirstAttr() {
		return follow(firstattribute);
	}

	/**
	 * Returns the first attribute node.
	 */
	public Attribute getFirstAttribute() {
		AttrNode a = getFirstAttr();
		if (a != null && !(a instanceof Attribute))
			a = a.getNextAttribute();
		return (Attribute)a;
	}

	/**
	 * Returns the first namespace declaration node.
	 */
	public NamespaceDecl getFirstNamespaceDecl() {
		return follow(firstnsdecl);
	}

	/**
	 * Returns the first child node.
	 */
	public XML getFirstChild() {
		return normalize(firstchild);
	}

	/**
	 * Returns the first child node, without normalizing.
	 */
	XML getRealFirstChild() {
		return follow(firstchild);
	}

	@Override
	void visitNormalizedBy(NodeVisitor v) {
		v.visit(this);
	}
	
	@Override
	<R,A> R visitNormalizedBy(NodeVisitorParameterized<R,A> v, A arg) {
		return v.visit(this, arg);
	}
	
	@Override
	void visitAnyBy(AnyNodeVisitor v) {
		v.visit(this);
	}

	/**
	 * Returns a copy of this node.
	 */
	Element copy() {
		return new Element(namespace, localname, firstattribute, firstchild, firstnsdecl, getNextSibling(), getOrigin());
	}
	
	/**
	 * Returns a copy of this node with the given next sibling.
	 */
	@Override
	public Element copy(XML nextsibling) {
		return new Element(namespace, localname, firstattribute, firstchild, firstnsdecl, nextsibling, getOrigin());
	}
	
	/**
	 * Returns a copy of this node with the given attributes, children, and next sibling.
	 */
	public Element copy(AttrNode firstattribute, XML firstchild, XML nextsibling) {
		return new Element(namespace, localname, firstattribute, firstchild, firstnsdecl, nextsibling, getOrigin());
	}
	
	/**
	 * Returns a copy of this node with the given first child and next sibling.
	 */
	Element copy(XML firstchild, XML nextsibling) {
		return new Element(namespace, localname, firstattribute, firstchild, firstnsdecl, nextsibling, getOrigin());
	}
	
	/**
	 * Returns a copy of this node with the given first attribute.
	 */
	Element copy(AttrNode firstattribute) {
		return new Element(namespace, localname, firstattribute, firstchild, firstnsdecl, getRealNextSibling(), getOrigin());
	}
	
	/**
	 * Returns a copy of this node with the given first namespace declarations.
	 */
	Element copy(NamespaceDecl firstnsdecl) {
		return new Element(namespace, localname, firstattribute, firstchild, firstnsdecl, getRealNextSibling(), getOrigin());
	}
	
	@Override
	public String toString() {
		return "[Element " + expanded(namespace, localname) + "]";
	}
	
	/**
	 * Returns the value of the attribute of the given name, or null if not present.
	 */
	@Override
	public String getAttribute(String localname) {
		return getAttribute(null, localname);
	}
	
	/**
	 * Returns the value of the attribute of the given name, or null if not present.
	 */
	@Override
	public String getAttribute(String namespace, String localname) {
		for (Attribute a = getFirstAttribute(); a != null; a = a.getNextAttribute())
			if (((a.getNamespace() == null && namespace == null) || a.getNamespace() != null && namespace != null && a.getNamespace().equals(namespace)) &&
					a.getLocalName().equals(localname))
				return a.getValue();
		return null;
	}
}
