package dk.brics.xact.analysis.dataflow;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import dk.brics.xact.analysis.flowgraph.Variable;

/**
 * Lattice element for {@link VariableAnalysis}.
 * Consists of a map from variables to analysis-specific value flow elements.
 */
public class VariableAnalysisElement<VariableElementType> {
	
	private Map<Variable,VariableElementType> vars = new LinkedHashMap<Variable,VariableElementType>();
	
	/**
	 * Constructs a new empty lattice element.
	 */
	public VariableAnalysisElement() {}
	
	/** 
	 * Adds or updates a mapping.
	 */
	public void put(Variable var, VariableElementType elem) {
		vars.put(var, elem);
	}
	
	/**
	 * Copies everything from the given lattice point. Note that the return
	 * value depends on the variable type element's {@link Object#equals(Object)} method
	 * to be implemented.
	 * @param other a lattice point
	 * @return true if this changed; false if they were already equal
	 */
	public boolean copyFrom(VariableAnalysisElement<VariableElementType> other) {
		if (vars.equals(other.vars))
			return false;
		vars.clear();
		vars.putAll(other.vars);
		return true;
	}
	
	/** 
	 * Gets the element associated with the given variable.
	 * Returns null if no explicit mapping exists.
	 */
	public VariableElementType get(Variable var) {
		return vars.get(var);
	}
	
	/** 
	 * Returns the variables for which explicit mappings exist.
	 */
	public Set<Variable> getVariables() {
		return vars.keySet();
	}
	
	/**
	 * Returns a string description of this lattice element.
	 */
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("(");
		boolean first = true;
		for (Variable v : vars.keySet()) {
			if (first)
				first = false;
			else
				sb.append(", ");
			sb.append(v + " -> " + vars.get(v));
		}
		sb.append(")");
		return sb.toString();
	}
}
