package dk.brics.xact.analysis.dataflow;

import dk.brics.xact.analysis.flowgraph.Statement;
import dk.brics.xact.analysis.flowgraph.statements.CastStm;
import dk.brics.xact.analysis.flowgraph.statements.CheckStm;
import dk.brics.xact.analysis.flowgraph.statements.ConcatStm;
import dk.brics.xact.analysis.flowgraph.statements.ConstStm;
import dk.brics.xact.analysis.flowgraph.statements.CopyStm;
import dk.brics.xact.analysis.flowgraph.statements.EmptyStm;
import dk.brics.xact.analysis.flowgraph.statements.GapifyStm;
import dk.brics.xact.analysis.flowgraph.statements.GetStm;
import dk.brics.xact.analysis.flowgraph.statements.InsertStm;
import dk.brics.xact.analysis.flowgraph.statements.NodeStm;
import dk.brics.xact.analysis.flowgraph.statements.PlugStm;
import dk.brics.xact.analysis.flowgraph.statements.RemoveStm;
import dk.brics.xact.analysis.flowgraph.statements.SetStm;
import dk.brics.xact.analysis.flowgraph.statements.UnknownStm;
import dk.brics.xact.analysis.flowgraph.statements.ValidateStm;

/**
 * Lattice and transfer functions for a {@link VariableAnalysis}.
 */
public interface VariableAnalysisInterface<VariableElementType> {

	/** 
	 * Creates a placeholder for a lattice element for a flow value, having the initial value.
	 */
	public VariableElementType newBottomElement();
	
	/** 
	 * Computes the least upper bound of two value lattice elements.
	 * @return <code>true</code> if the new value of <code>dest</code> is
	 *         different from the old, <code>false</code> otherwise.  
	 */
	public boolean merge(VariableElementType source, VariableElementType dest);

	/**
	 * Invoked after each assignment transfer function on the resulting value.
	 */
	public void assign(VariableElementType value, Statement s);

	/**
	 * Transfer function for a {@link CastStm}.
	 */
	public VariableElementType transferCast(CastStm s, VariableElementType base);

	/**
	 * Transfer function for a {@link CheckStm}.
	 */
	public void transferCheck(CheckStm s, VariableElementType base);

	/**
	 * Transfer function for a {@link ConcatStm}.
	 */
	public VariableElementType transferConcat(ConcatStm s, VariableElementType xmlsource);

	/**
	 * Transfer function for a {@link ConstStm}.
	 */
	public VariableElementType transferConst(ConstStm s);

	/**
	 * Transfer function for a {@link CopyStm}.
	 */
	public VariableElementType transferCopy(CopyStm s, VariableElementType base,
			VariableElementType firstattr, VariableElementType firstchild, VariableElementType nextnode);

	/**
	 * Transfer function for an {@link EmptyStm}.
	 */
	public VariableElementType transferEmpty(EmptyStm s);

	/**
	 * Transfer function for a {@link GapifyStm}.
	 */
	public VariableElementType transferGapify(GapifyStm s, VariableElementType base);

	/**
	 * Transfer function for a {@link GetStm}.
	 */
	public VariableElementType transferGet(GetStm s, VariableElementType base);

	/**
	 * Transfer function for an {@link InsertStm}.
	 */
	public VariableElementType transferInsert(InsertStm s, VariableElementType base, VariableElementType xmlsrc);

	/**
	 * Transfer function for a {@link NodeStm}.
	 */
	public VariableElementType transferNode(NodeStm s, 
			VariableElementType firstattr, VariableElementType firstchild, VariableElementType nextnode);

	/**
	 * Transfer function for a {@link PlugStm}.
	 */
	public VariableElementType transferPlug(PlugStm s, VariableElementType variable, VariableElementType xmlsrc);

	/**
	 * Transfer function for a {@link RemoveStm}.
	 */
	public VariableElementType transferRemove(RemoveStm s, VariableElementType base);

	/**
	 * Transfer function for a {@link SetStm}.
	 */
	public VariableElementType transferSet(SetStm s, VariableElementType base, VariableElementType xmlsrc);

	/**
	 * Transfer function for an {@link UnknownStm}.
	 */
	public VariableElementType transferUnknown(UnknownStm s);

	/**
	 * Transfer function for a {@link ValidateStm}.
	 */
	public VariableElementType transferValidate(ValidateStm s, VariableElementType base);

}
