package dk.brics.xact.analysis.flowgraph;

import java.util.*;

/** 
 * Generic directed multigraph.
 */
public class Graph<Nodetype,Edgetype> { 
	
	private Set<Nodetype> nodes = new LinkedHashSet<Nodetype>();    
	
	private Map<Nodetype,Set<Edge<Nodetype,Edgetype>>> inedges = new LinkedHashMap<Nodetype,Set<Edge<Nodetype,Edgetype>>>();  
	
	private Map<Nodetype,Set<Edge<Nodetype,Edgetype>>> outedges = new LinkedHashMap<Nodetype,Set<Edge<Nodetype,Edgetype>>>(); 
	
	private Set<Nodetype> entries = new LinkedHashSet<Nodetype>();  
	
	/** 
	 * Constructs an empty graph.
	 */
	public Graph() {}
	
	private void checkNodeExists(Nodetype n) {
		if (!containsNode(n))
			throw new NoSuchElementException("No such node: " + n);
		
	}
	
	/** 
	 * Adds all nodes and edges from another graph to this one.
	 */
	public void addAll(Graph<Nodetype,Edgetype> other) {
		nodes.addAll(other.nodes);
		inedges.putAll(other.inedges);
		outedges.putAll(other.outedges);
		entries.addAll(other.entries);
	}
	
	/** 
	 * Adds the given node to the graph. If the node was already in the graph then nothing happens.
	 */
	public void addNode(Nodetype node) {
		// if the node was already there, avoid overwriting the edge sets
		if (nodes.add(node)) {
			inedges.put(node, new LinkedHashSet<Edge<Nodetype,Edgetype>>());
			outedges.put(node, new LinkedHashSet<Edge<Nodetype,Edgetype>>());
		}
	}
	
	/** 
	 * Removes the given node from the graph.
	 * All edges to or from this node are also removed.
	 * @exception NoSuchElementException if the node does not exist in the graph
	 */
	public void removeNode(Nodetype node) {
		checkNodeExists(node);
		Iterator<Edge<Nodetype,Edgetype>> outs = getOutEdges(node).iterator();
		while (outs.hasNext()) {
			Edge<Nodetype,Edgetype> edge = outs.next();
			outs.remove();
		    inedges.get(edge.getTo()).remove(edge);
		}
		Iterator<Edge<Nodetype,Edgetype>> ins = getInEdges(node).iterator();
		while (ins.hasNext()) {
			Edge<Nodetype,Edgetype> edge = ins.next();
			ins.remove();
			outedges.get(edge.getFrom()).remove(edge);
		}
		nodes.remove(node);
		if (entries.contains(node)) 
			entries.remove(node);
	}
	
	/** 
	 * Adds an edge between two nodes in the graph.
	 * @exception NoSuchElementException if either of the nodes does not exist in the graph
	 */
	public void addEdge(Nodetype from, Nodetype to, Edgetype data) { 
		addEdge(new Edge<Nodetype,Edgetype>(from, to, data));
	}
	
	private void addEdge(Edge<Nodetype,Edgetype> edge) { 
		checkNodeExists(edge.getFrom());
		checkNodeExists(edge.getTo());
		outedges.get(edge.getFrom()).add(edge);
		inedges.get(edge.getTo()).add(edge);
	}
	
	/** 
	 * Removes an existing edge between two nodes in the graph.
	 * @exception NoSuchElementException if the edge does not exist
	 */
	public void removeEdge(Edge<Nodetype,Edgetype> edge) { 
		outedges.get(edge.getFrom()).remove(edge);
		inedges.get(edge.getTo()).remove(edge);
	}
	
	/**
	 * Removes all outgoing edges from the given node.
	 */
	public void removeOutEdges(Nodetype node) {
		Set<Edge<Nodetype,Edgetype>> outs = getOutEdges(node);
		Iterator<Edge<Nodetype,Edgetype>> oi = outs.iterator();
		while (oi.hasNext()) {
			Edge<Nodetype,Edgetype> edge = oi.next();
			oi.remove();
		    inedges.get(edge.getTo()).remove(edge);
		}
		outs.clear();
	}
	/**
	 * Removes all ingoing edges from the given node.
	 */
	public void removeInEdges(Nodetype node) {
		Set<Edge<Nodetype,Edgetype>> ins = getInEdges(node);
		Iterator<Edge<Nodetype,Edgetype>> ii = ins.iterator();
		while (ii.hasNext()) {
			Edge<Nodetype,Edgetype> edge = ii.next();
			ii.remove();
		    outedges.get(edge.getFrom()).remove(edge);
		}
		ins.clear();
	}
	
	/**
	 * Returns the total number of edges.
	 */
	public int getNumberOfEdges() {
		int noe = 0;
		for (Nodetype node : getNodes())
			noe += getOutEdges(node).size();
		return noe;
	}
	
	/** 
	 * Marks an existing node as an entry node of the graph.
	 * @exception NoSuchElementException if the node does not exist
	 */
	public void addEntry(Nodetype node) { 
		checkNodeExists(node);
		entries.add(node);
	}
	
	/** 
	 * Removes an entry node of the graph.
	 * @exception NoSuchElementException if the node does not exist
	 */
	public void removeEntry(Nodetype node) { 
		if (!entries.contains(node))
			throw new NoSuchElementException("No such entry node: " + node);
		entries.remove(node);
	}
	
	/**
	 * Returns the set of entry points in the graph. It is
	 * preferrable to use {@link #addEntry(Object)} and {@link #removeEntry(Object)}
	 * instead of modifying the set directly.
	 * @return mutable reference to the set of entities
	 */
	public Set<Nodetype> getEntries() { 
		return entries;
	}
	
	/** 
	 * Checks whether the graph contains the given node.
	 */
	public boolean containsNode(Nodetype node) { 
		return nodes.contains(node);
	}
	
	/**
	 * Returns the set of outgoing edges from the specified node.
	 * @param node a node in the graph
	 * @return a set of edges. Should not be modified.
	 */
	public Set<Edge<Nodetype,Edgetype>> getOutEdges(Nodetype node) { 
		checkNodeExists(node);
		if (outedges.containsKey(node))
			return outedges.get(node);
		else
			return new LinkedHashSet<Edge<Nodetype,Edgetype>>();
	}
	
	/**
	 * Returns the set of ingoing edges to the specified node.
	 * @param node a node in the graph
	 * @return a set of edges. Should not be modified.
	 */
	public Set<Edge<Nodetype,Edgetype>> getInEdges(Nodetype node) { 
		checkNodeExists(node);
		if (inedges.containsKey(node))
			return inedges.get(node);
		else
			return new LinkedHashSet<Edge<Nodetype,Edgetype>>();
	}
	
	/**
	 * Returns the set of all nodes in the graph.
	 * @return reference to the set of nodes. Should not be modified.
	 */
	public Set<Nodetype> getNodes() { 
		return nodes;
	}
	
	/**
	 * Removes all edges.
	 */
	public void clearEdges() {
		for (Set<Edge<Nodetype,Edgetype>> s : inedges.values())
			s.clear();
		for (Set<Edge<Nodetype,Edgetype>> s : outedges.values())
			s.clear();
	}
}
