package dk.brics.xact.analysis.util;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * Maintains a set of disjoint sets (called components). A node belongs to exactly one component,
 * and components can be merged into larger components. Each component is associated with
 * some user data that can be accessed with {@link #getData(Object)}.
 * 
 * @param <Key>
 * @param <NodeData>
 */
public abstract class UnionFindForest<Key,NodeData> {
	private Map<Key,Node> map = new HashMap<Key,Node>();
	
	private class Node {
		Key key;
		Node parent;
		NodeData data;
		int rank;
		Node(Key key, NodeData data) {
			this.key = key;
			this.parent = this;
			this.data = data;
			this.rank = 1;
		}
	}
	
	/**
	 * Merges the data associated with two components when they are merged into one. 
	 * Both arguments may be modified if wanted and the returned value may either be 
	 * one of the arguments or a new instance.
	 * @param a data from one component
	 * @param b data from another component
	 * @return new node for their joined component
	 */
	protected abstract NodeData mergeNodeData(NodeData a, NodeData b);
	protected abstract NodeData defaultNodeData();
	
	private Node getrep(Node node) {
		if (node.parent == node) {
			assert node.data != null;
			return node;
		} else {
			return node.parent = getrep(node.parent);
		}
	}
	private Node getrepOrAdd(Key key) {
		Node node = map.get(key);
		if (node == null) {
			node = new Node(key, defaultNodeData());
			map.put(key, node);
			return node;
		} else {
			return getrep(node);
		}
	}
	public NodeData merge(Key a, Key b) {
		Node na = getrepOrAdd(a);
		Node nb = getrepOrAdd(b);
		if (na == nb) // already merged
			return na.data;
		if (na.rank < nb.rank) {
			na.parent = nb;
			nb.data = mergeNodeData(na.data, nb.data);
			na.data = null;
			return nb.data;
		} else {
			if (na.rank == nb.rank) {
				na.rank++;
			}
			nb.parent = na;
			na.data = mergeNodeData(na.data, nb.data);
			nb.data = null;
			return na.data;
		}
	}
	public boolean containsKey(Key key) {
		return map.containsKey(key);
	}
	/**
	 * Returns the data associated with the specified key's component.
	 * The data may be modified if wanted.
	 */
	public NodeData getData(Key key) {
		return getrepOrAdd(key).data;
	}
	/**
	 * Sets the data associated with the specified key's component.
	 */
	public void setData(Key key, NodeData data) {
		getrepOrAdd(key).data = data;
	}
	/**
	 * Returns the key representing the given key's component. This
	 * is the same for all keys in the same component.
	 */
	public Key getRepresentativeKey(Key key) {
		return getrepOrAdd(key).key;
	}
	public Set<Key> getKeys() {
		return map.keySet();
	}
}
