package dk.brics.xact.analysis.xmlgraph;

import java.util.Set;

import dk.brics.xmlgraph.AttributeNode;
import dk.brics.xmlgraph.ChoiceNode;
import dk.brics.xmlgraph.ElementNode;
import dk.brics.xmlgraph.InterleaveNode;
import dk.brics.xmlgraph.Node;
import dk.brics.xmlgraph.NodeProcessor;
import dk.brics.xmlgraph.SequenceNode;
import dk.brics.xmlgraph.XMLGraph;

/**
 * Determines which elements may occur as first root in some unfolding of
 * an XML graph.
 * <p/>
 * More specifically, a node <i>N</i> may be marked "first root=YES" if the following holds:
 * <blockquote>
 * For every unfolding of the XML graph and for every occurence <i>X</i> of an unfolding of <i>N</i> inside
 * the unfolded document, the first node in <i>X</i> is the first root in the unfolded document.
 * </blockquote>
 * Conversely, a node <i>N</i> may be marked "first root=NO" if the following holds:
 * <blockquote>
 * For every unfolding of the XML graph and for every occurence <i>X</i> of an unfolding of <i>N</i> inside
 * the unfolded document, the first node in <i>X</i> is <i>not</i> the first root in the unfolded document.
 * </blockquote>
 * If both properties are dissatisfied, then "first root=MAYBE" is the correct answer. 
 * <p/>
 * This definition means that a sequence node can be marked "first root" even though it unfolds to a sequence of
 * nodes, of which only the first is the first root. For example:
 * <pre>
 *  (root)
 *  SequenceNode (first root = YES)
 *    |
 *    +-> ChoiceNode (first root = YES)
 *    |       |
 *    |       +-> ElementNode (first root = YES)   
 *    |
 *    +-> ChoiceNode (first root = NO)
 *            |
 *            +-> ElementNode (first root = NO)   
 * </pre>
 * An example where "first root" is MAYBE would be:
 * <pre>
 *          (root)
 *       SequenceNode (first root = YES)
 *        /         \
 *  ChoiceNode    ChoiceNode
 *        \         /
 *        ElementNode (first root = MAYBE)
 * </pre>
 * <p/>
 * The analysis can be configured to consider attributes, elements, text nodes and gaps
 * specially through its underlying emptiness analysis. Properly configured, it can
 * answer questions such as <em>"which nodes can occur as the first element root"</em>.
 */
public class FirstRootAnalysis extends ForwardsXGAnalyzer<EBooleanLattice> {
    
    private Emptiness emptiness;
    private boolean allowText;
    
    public FirstRootAnalysis(XMLGraph xg, Emptiness emptiness, boolean allowText) {
        super(xg);
        this.emptiness = emptiness;
        this.allowText = allowText;
        doAnalysis();
    }
    
    @Override
    protected EBooleanLattice bottom() {
        return EBooleanLattice.BOTTOM;
    }

    @Override
    protected EBooleanLattice initial(Node node) {
        EPresence e = emptiness.get(node);
        
        // if no unfolding, use bottom
        if (e == EPresence.BOTTOM)
            return EBooleanLattice.BOTTOM;
        
        // elements and attributes cannot unfold to an empty sequence
        if (e == EPresence.EMPTY && !allowText)
            return EBooleanLattice.BOTTOM;
        
        return EBooleanLattice.YES;
    }
    
    @Override
    protected Set<Integer> initialNodes() {
        return getXmlGraph().getRoots();
    }
    
    @Override
    protected void transfer(EBooleanLattice src, Node node) {
        processor.src = src;
        node.process(processor);
    }
    
    private Processor processor = new Processor();
    private final class Processor extends NodeProcessor<Object> {
        EBooleanLattice src;
        @Override
        public Object process(ElementNode n) {
            merge(n.getContent(), EBooleanLattice.NO);
            return this;
        }
        @Override
        public Object process(AttributeNode n) {
            merge(n.getContent(), EBooleanLattice.NO);
            return this;
        }
        @Override
        public Object process(SequenceNode n) {
            EPresence p = EPresence.EMPTY;
            for (int child : n.getContents()) {
                if (p == EPresence.BOTTOM)
                    break;
                if (p.definitelyEmpty()) {
                    merge(child, src);
                } else if (p.definitelyNonEmpty()) {
                    merge(child, EBooleanLattice.NO);
                } else {
                    merge(child, EBooleanLattice.MAYBE);
                }
                p = p.concat(emptiness.get(child));
            }
            return this;
        }
        @Override
        public Object process(ChoiceNode n) {
            for (int child : n.getContents()) {
                merge(child, src);
            }
            return this;
        }
        @Override
        public Object process(InterleaveNode n) {
            if (n.getContents().size() == 1) {
                merge(n.getContents().iterator().next(), src);
            } else {
                for (int child : n.getContents()) {
                    merge(child, EBooleanLattice.MAYBE);
                }
            }
            return this;
        }
        
        private void merge(int target, EBooleanLattice val) {
            put(target, get(target).leastUpperBound(val));
        }
    }
    
    
}
