package dk.brics.xact.operations;

import dk.brics.xact.AttrNode;
import dk.brics.xact.Element;
import dk.brics.xact.NamespaceDecl;
import dk.brics.xact.XML;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

/**
 * Manages namespace prefixes during construction of ordinary XML output from XML templates.
 */
class NamespacePrefixTracker {

    private static final String XHTML_NAMESPACE = "http://www.w3.org/1999/xhtml";

    private final Map<String, Stack<String>> ns2prefix = new HashMap<String, Stack<String>>();

    private final Map<String, String> suggestedprefix = new HashMap<String, String>();

    private final Map<String, Stack<String>> prefix2ns = new HashMap<String, Stack<String>>();

    /**
     * Constructs a new namespace prefix tracker.
     */
    public NamespacePrefixTracker() {
        addDecl("", "");
        addDecl("xml", "http://www.w3.org/XML/1998/namespace");
        for (Map.Entry<String, String> me : XML.getNamespaceMap().entrySet()) {
            suggestedprefix.put(me.getValue(), me.getKey());
        }
        for (Map.Entry<String, String> me : XML.getThreadNamespaceMap().entrySet()) {
            suggestedprefix.put(me.getValue(), me.getKey());
        }
        suggestedprefix.put("", "");
        suggestedprefix.put(XHTML_NAMESPACE, "");
    }

    /**
     * Adds the list of namespaces declared by the element to the namespace tracker.
     * Even if declared on this element, declarations will be ignored (not pushed) if a
     * the same namespace has already been pushed with the same prefix in an ancestor
     * and this declaration is still active.
     *
     * @return all the declarations that have been added
     */
    public Map<String, String> pushNamespaceDeclarations(Element n) {
        Map<String, String> nsdecls = new HashMap<String, String>();
        for (NamespaceDecl nd = n.getFirstNamespaceDecl(); nd != null; nd = nd.getNext()) {
            String prefix = nd.getPrefix();
            String namespace = nd.getNamespace();
            Stack<String> prefixStack = ns2prefix.get(namespace);
            //Skip namespaces that have already been declared with the same prefix
            if (prefixStack != null && !prefixStack.isEmpty() && prefixStack.peek().equals(prefix)) {
                continue;
            }
            nsdecls.put(prefix, namespace);
            addDecl(prefix, namespace);
        }
        ensureDeclared(n.getNamespace(), nsdecls);
        for (AttrNode a = n.getFirstAttr(); a != null; a = a.getNextAttr()) {
            if (a.getNamespace() != null) {
                ensureDeclared(a.getNamespace(), nsdecls);
            }
        }
        return nsdecls;
    }

    /**
     * Ensures that the given namespace is declared.
     *
     * @param nsdecls if a new declaration is needed, it is also added here
     */
    public void ensureDeclared(String namespace, Map<String, String> nsdecls) {
        String prefix = suggestedprefix.get(namespace);
        String currentns = null;
        if (prefix != null && prefix2ns.get(prefix) != null && !prefix2ns.get(prefix).isEmpty()) {
            currentns = prefix2ns.get(prefix).peek();
        }
        if (!ns2prefix.containsKey(namespace) || (prefix != null && !namespace.equals(currentns))) {
            if (prefix == null) {
                prefix = "n" + (ns2prefix.size() - 1);
            }
            nsdecls.put(prefix, namespace);
            addDecl(prefix, namespace);
        }
    }

    /**
     * Adds a namespace declaration.
     */
    public void addDecl(String prefix, String namespace) {
        Stack<String> prefixstack = ns2prefix.get(namespace);
        if (prefixstack == null) {
            prefixstack = new Stack<String>();
            ns2prefix.put(namespace, prefixstack);
        }
        prefixstack.push(prefix);
        Stack<String> nsstack = prefix2ns.get(prefix);
        if (nsstack == null) {
            nsstack = new Stack<String>();
            prefix2ns.put(prefix, nsstack);
        }
        nsstack.push(namespace);
    }

    /**
     * Pops the given namespace declarations.
     */
    public void popNamespaceDeclarations(Map<String, String> nsdecls) {
        for (String ns : nsdecls.values()) {
            String prefix = ns2prefix.get(ns).pop();
            if (ns2prefix.get(ns).isEmpty()) {
                ns2prefix.remove(ns);
            }
            prefix2ns.get(prefix).pop();
            if (prefix2ns.get(prefix).isEmpty()) {
                prefix2ns.remove(prefix);
            }
        }
    }

    /**
     * Returns the prefix for the given namespace.
     */
    public Stack<String> getPrefix(String n) {
        return ns2prefix.get(n);
    }
}
