package dk.brics.xact.operations;

import dk.brics.xact.Element;

/**
 * Specifies how to indent an XML template or document. The default implementation
 * {@link #NEVER} will disable all indentation and is recommended for production use.
 * Other available implementations are {@link #XHTML} and {@link #ALWAYS}, though
 * one can provide his own implementation if necessary.
 */
public interface XMLIndentation {
	enum IndentType {
		BLOCK,
		INLINE_BLOCK,
		NONE
	}
	
    /**
     * Decide whether indentation should be increased when printing
     * the children of the specified element.
     * <p/>
     * In XHTML, elements such as <tt>&lt;a&gt;</tt> and <tt>&lt;u&gt;</tt> should not be indented
     * because they are sensitive to leading and trailing whitespace.
     * @param element an XML element about to be printed
     * @return true if indentation should be enabled
     */
    IndentType getIndentType(Element element);
    
    /**
     * Returns true if the specified element is sensitive to repeated whitespace.
     * <p/>
     * In XHTML, elements such as <tt>&lt;pre&gt;</tt> and <tt>&lt;textarea&gt;</tt> are sensitive
     * to repeated whitespace, though any element can be made sensitive through CSS.
     * @param element an XML element about to be printed
     * @return true if element is whitespace-sensitive
     */
    boolean isWhitespaceSensitive(Element element);
    
    /**
     * Returns the string used to represent linebreaks. Typically <tt>\r\n</tt>, <tt>\r</tt> or <tt>\n</tt>.
     * @return a string
     */
    String getLineBreak();
    
    /**
     * Returns the indentation string to prefix every line for each level of indentation.
     * @return indentation string
     */
    String getIndentation();
    
    /**
     * Default implementation that ensures no indentation will be performed.
     */
    public static final XMLIndentation NEVER = new XMLIndentation() {
        public boolean isWhitespaceSensitive(Element element) {
            return true;
        }
        public IndentType getIndentType(Element element) {
        	return IndentType.NONE;
        }
        public String getLineBreak() {
            return "";
        }
        public String getIndentation() {
            return "";
        }
    };
    /**
     * Implementation that indents all elements using four whitespaces per level
     * and uses the system property <tt>line.separator</tt> as line separator.
     */
    public static final XMLIndentation ALWAYS = new XMLIndentation() {
        public String getIndentation() {
            return "    ";
        }
        public String getLineBreak() {
            return System.getProperty("line.separator");
        }
        public boolean isWhitespaceSensitive(Element element) {
            return false;
        }
        public IndentType getIndentType(Element element) {
        	return IndentType.BLOCK;
        }
    };
    /**
     * An instance of of the {@link XHTMLIndentation} class, provided here for convenience.
     */
    public static final XHTMLIndentation XHTML = new XHTMLIndentation();
}
